# -*- coding: utf-8 -*-
"""
@author: P.Cizek
"""
import sys
import oscilator_network as osc
import oscilator_constants as const
import hexapod_vrep.robot as rob
import hexapod_vrep.constants as rc
import time
import math
import thread

def goto(goal_x,goal_y,goal_phi):
	global left_t
	global right_t
	l = 1
	r = 1
	adjust_heading = False
	done = False
	
	while done == False:
		#get robot position and orientation in the global reference frame
		position = robot.get_robot_position()
		orientation = robot.get_robot_orientation() - math.pi
	
		#get vector to target
		dx = position[0] - goal_x 
		dy = position[1] - goal_y
		
		#get orientation to the target
		phi = math.atan2(dy,dx);
		if phi < 0:
			phi+=2*math.pi
			
		#get the difference between orientation angles
		#if the robot is in the close vicinity of the goal position, we exchange the direction to the target for the desired orientation of the robot
		if adjust_heading == False:
			dphi = phi - orientation
		else:
			dphi = goal_phi - orientation
		
		dphi = (dphi + math.pi) % (2*math.pi) - math.pi	#make the difference between angles to be within <-pi,pi> , in such a way, it is possible to select a shorter turn to target
		
		#get distance to the target
		dist = math.sqrt(dx*dx + dy*dy) 
		
		#debug print to see the distance to the target, robot orientation, difference between orientation angles and action command
		print(str(dist) + ' ' + str(orientation) + " " + str(dphi)+ " " + str(left_t) + " "+ str(right_t))
		
		#rotate the robot in the direction of walking if the angle differencee is big
		if math.fabs(dphi) > math.pi/8:
			if dphi > 0:	#turn either to left or right dependng on to which side it is closer to fix the heading
				l = 1	#left steering
				r = -1	#right steering
			else:
				l = -1	#left steering
				r = 1	#right steering
		#else the heading is approximately correct so go straight to the target
		else:
			if adjust_heading == False:
				l = -1*dist*2	#left steering
				r = -1*dist*2	#right steering
				l = max(-1, min(l, 1))	#limit interval for steering to <-1,1>
				r = max(-1, min(r, 1))	#limit interval for steering to <-1,1>
		
		#if the robot is at the goal position start turning to correct the heading
		if math.fabs(dist) < 0.1:
			adjust_heading = True
		
		#if the robot is at the goal position with a correct heading, stop
		if math.fabs(dist) < 0.1 and math.fabs(dphi) < math.pi/8:
			done = True
		
		#setup driving parameters = commmand the locomotion thread
		Lock.acquire()
		left_t = l
		right_t = r
		Lock.release()

def locomotion():
	#cpg oscillator instantiation
	cpg = osc.OscilatorNetwork()
	cpg.change_gait(const.TRIPOD_GAIT_WEIGHTS)
	
	coxa_angles= [0,0,0,0,0,0]
	cycle_length = [0,0,0,0,0,0]
	
	#steering commands
	global left_t
	global right_t
	left = left_t
	right = right_t
	
	#main locomotion control loop
	global stop
	while stop == False:
		
		# steering - acquire left and right steering speeds
		Lock.acquire()
		left = left_t
		right = right_t
		Lock.release()
		coxa_dir = [left,left,left,right,right,right]	#set directions for individual legs

		#next step of CPG
		cycle = cpg.oscilate_all_CPGs()
		#read the state of the network
		data = cpg.get_last_values()

		#reset coxa angles if new cycle is detected
		for i in range(0,6):
			cycle_length[i] += 1;
			if cycle[i] == True:
				coxa_angles[i]= -((cycle_length[i]-2)/2)*rc.COXA_MAX
				cycle_length[i] = 0

		#calculate individual joint angles for each of six legs
		for i in range(0,6):
			femur_val = rc.FEMUR_MAX*data[i] #calculation of femur angle
			if femur_val < 0:
				coxa_angles[i] += coxa_dir[i]*rc.COXA_MAX	#calculation of coxa angle -> stride phase
				femur_val *= 0.1
			else:
				coxa_angles[i] -= coxa_dir[i]*rc.COXA_MAX	#calculation of coxa angle -> stance phase
			
			coxa_val = coxa_angles[i]
			tibia_val = -rc.TIBIA_MAX*data[i]	#calculation of tibia angle
			
			#set position of each servo
			robot.set_servo_position(rc.COXA_SERVOS[i], rc.SIGN_SERVOS[i]*coxa_val+rc.COXA_OFFSETS[i])
			robot.set_servo_position(rc.FEMUR_SERVOS[i],rc.SIGN_SERVOS[i]*femur_val+rc.FEMUR_OFFSETS[i])
			robot.set_servo_position(rc.TIBIA_SERVOS[i],rc.SIGN_SERVOS[i]*tibia_val+rc.TIBIA_OFFSETS[i])
			
			#simulation step time
			time.sleep(0.001)
	
	thread.exit_thread()
	return

if __name__ == "__main__":
	#instantiate robot
	robot = rob.Robot()

	#define global variables used for the robot control
	Lock = thread.allocate_lock()	#mutex for access to turn commands
	l = 1	#left turning speed from interval <-1,1>
	r = 1	#right turning speed from interval <-1,1>
	global left_t
	global right_t
	left_t = l
	right_t = r

	global stop	#global variable to terminate the locomotion thread
	stop = False

	#join walking thread - comply to the hierarchical architecture where robot cintrol is on the bottom level
	try:
		thread1 = thread.start_new_thread(locomotion, ())
	except:
		print("Error: unable to start new thread")

	#walk around a rectangle
	goto(1,0,math.pi/2)
	print("goal reached")
	goto(1,1,math.pi)
	print("goal reached")
	goto(0,1,-math.pi/2)
	print("goal reached")
	goto(0,0,-math.pi)
	print("goal reached")
	stop = True

	#stop the simulation and disconnect from simulator
	robot.stop_simulation()
	robot.disconnect_simulator()
	sys.exit()